require 'installer/action'
require 'installer/errors'

class Installer::Planner::Step::CompatibilityCheck < Installer::Planner::Step
  def initialize(infra, data={}, pre=nil, post=nil)
    raise 'No package dir provided' unless data[:packages_dir]
    @packages_dir = data[:packages_dir]
    super
  end

  def get_available_platforms
    Dir.glob("#{@packages_dir}/*").select { |f| File.directory?(f) }.map { |f| File.basename(f) }
  end

  def perform_step(&execute)
    os_data = Hash.new {|h,k| h[k] = {}}

    # Identify the Installer being used
    available_platforms = get_available_platforms

    @infra.hosts.each do |hostname, host|
      # Ensure PE is not already installed
      peinstalled = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "ls /opt/puppet/bin/puppet")
      yield peinstalled

      if peinstalled.success?
        peversion = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /opt/puppet/pe_version")
        yield peversion

        if peversion.success?
          error(hostname, @t.step.compatibility_check.pe_exists.known_version(version: peversion.results[hostname][:stdout].strip))
        else
          error(hostname, @t.step.compatibility_check.pe_exists.unknown_version)
        end
      end

      # First try identifying using lsb_release.  This takes care of Ubuntu (lsb-release is part of ubuntu-minimal).
      lsb = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "lsb_release -icr")
      yield lsb

      if lsb.success?
        lsb_data = {}
        lsb.results[hostname][:stdout].each_line do |l|
          k, v = l.split(':')
          if k && v
            lsb_data[k.strip] = v.strip
          end
        end

        distribution = lsb_data['Distributor ID'].downcase.gsub(/\s+/, "")
        distribution = case distribution
        when /redhatenterpriseserver|redhatenterpriseclient|redhatenterpriseas|redhatenterprisees|enterpriseenterpriseserver|redhatenterpriseworkstation|redhatenterprisecomputenode|oracleserver/
          :rhel
        when /enterprise.*/
          :centos
        when /scientific|scientifics|scientificsl/
          :rhel
        when /amazonami/
          :amazon
        when /suselinux/
          :sles
        else
          distribution.to_sym
        end

        release = lsb_data['Release'].gsub(/\s+/, "")
        release = case distribution
        when :centos, :rhel
          release.split('.')[0]
        when :debian
          if release == 'testing'
            '7'
          else
            release.split('.')[0]
          end
        else
          release
        end

      end

      # Check for Redhat
      if !distribution && !release
        redhat_release = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/redhat-release")
        yield redhat_release

        if redhat_release.success?
          stdout = redhat_release.results[hostname][:stdout]

          distribution = case stdout
          when /red hat enterprise/im
            :rhel
          when /centos/im
            :centos
          when /scientific/im
            :rhel
          end

          release = /.* release ([[:digit:]]).*/.match(stdout)[1]
        end
      end

      # Check for Cumulus
      if !distribution && !release
        os_release = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/os-release")
        yield os_release

        if os_release.success?
          stdout = os_release.results[hostname][:stdout]

          if /Cumulus Linux/m.match(stdout)
            distribution = :cumulus
          end
          if distribution
            release = /VERSION_ID=(\d+\.\d)/.match(stdout)[1]
          end
        end
      end

      # Check for EOS
      if !distribution && !release
        eos_release = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/Eos-release")
        yield eos_release

        if eos_release.success?
          stdout = eos_release.results[hostname][:stdout]

          if /Arista Networks EOS/m.match(stdout)
            distribution = :eos
          end

          release = /^Arista Networks EOS v*(.*)\..*$/.match(stdout)[1]
        end
      end

      # Check for Debian
      if !distribution && !release
        debian_version = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/debian_version")
        yield debian_version

        if debian_version.success?
          stdout = debian_version.results[hostname][:stdout]
          distribution = :debian

          release = case stdout
          when /^[[:digit:]]/
            stdout.split('.')[0]
          when /^wheezy/.match(stdout)
            "7"
          end
        end
      end

      # Check for SuSE
      if !distribution && !release
        suse_release = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/SuSE-release")
        yield suse_release

        if suse_release.success?
          stdout = suse_release.results[hostname][:stdout]

          if /Enterprise Server/.match(stdout)
            distribution = :sles
            release = /^VERSION = (\d*)/m.match(stdout)[1]
          end
        end
      end

      # Check for Amazon 6, or fail
      if !distribution && !release
        system_release = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "cat /etc/system-release")
        yield system_release

        if system_release.success?
          stdout = system_release.results[hostname][:stdout]

          if /amazon linux/im.match(stdout)
            distribution = :amazon
            # How is this safe to assume?
            release = "6"
          else
            error(host, @t.step.compatibility_check.unsupported_platform(platform: stdout))
            next
          end
        end
      end

      # Check for Solaris
      if !distribution && !release
        uname = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "uname -s")
        yield uname

        if uname.success?
          distribution = case uname.results[hostname][:stdout].strip
          when "SunOS"
            unamer = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "uname -r")
            yield unamer

            if unamer.success?
              release = unamer.results[hostname][:stdout].split('.')[0]
            end

            :solaris
          when "AIX"
            oslevel = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "oslevel")
            yield oslevel

            if oslevel.success?
              release = oslevel.results[hostname][:stdout].split('.')[0..1].join('.')
            end

            :aix
          end
        end
      end

      # If we haven't found the distribution or release by this point, it's definitely not supported.
      unless distribution && release
        error(hostname, @t.step.compatibility_check.unknown_platform)
      end

      # Architecture
      unamem = Installer::Action::Execute.new(:host => host, :sudo => true, :cmd => "uname -m")
      yield unamem

      if unamem.success?
        architecture = unamem.results[hostname][:stdout].strip

        architecture = case architecture
        when 'i686'
          'i386'
        when 'ppc'
          'powerpc'
        when 'x86_64'
          [:ubuntu, :debian].include?(distribution) ? 'amd64' : 'x86_64'
        else
          architecture
        end
      end

      unless architecture
        error(hostname, @t.step.compatibility_check.unknown_architecture)
      end

      unless distribution && release && architecture
        next
      end

      # Tag
      os_data[hostname][:distribution] = distribution
      os_data[hostname][:release] = release
      os_data[hostname][:architecture] = architecture
      os_data[hostname][:tag] = case distribution
      when :rhel, :centos, :amazon
        "el-#{release}-#{architecture}"
      else
        "#{distribution}-#{release}-#{architecture}"
      end
    end

    platform_tags = os_data.values.collect { |d| d[:tag] }.uniq

    if platform_tags.size > 1
      error("localhost", @t.step.compatibility_check.multiple_platforms(platforms: platform_tags.join(', ')))
    elsif platform_tags.empty?
      error("localhost", @t.step.compatibility_check.no_platforms_detected)
    end

    platform_tag = platform_tags.first

    if platform_tag && !available_platforms.include?(platform_tag.to_s)
      error("localhost", @t.step.compatibility_check.platform_not_available(platform: platform_tag, available: available_platforms.join(', ')) )
    end

    os_data
  end

  def name
    :compatibility_check
  end

  def description
    @t.step.compatibility_check.description(hostname: @infra.installer_hostname, hosts: @infra.hosts.keys.join(', '))
  end
end
