require 'installer/action'
require 'installer/errors'

class Installer::Planner::Step::InstallPE < Installer::Planner::Step
  def initialize(infra, data={}, pre=nil, post=nil)
    super

    @host = @infra.hosts[data[:host]] if data[:host]
    @stop_on_fail = true

    raise 'Undefined remote logging directory' unless data[:log_dir]
    @log_dir = data[:log_dir]

    raise 'Undefined local installer directory' unless data[:local_installer_dir]
    @local_installer_dir = data[:local_installer_dir]

    raise 'Undefined configuration directory' unless data[:etc_dir]
    @etc_dir = data[:etc_dir]

    raise 'Undefined local answer directory' unless data[:local_answer_dir]
    @local_answer_dir = data[:local_answer_dir]
  end

  # Call from executor with a block to handle the actions.
  def perform_step(&execution_block)
    mktmp = Installer::Action::Execute.new(:hosts => [@host], :cmd => "mktemp -d -t pe-installer-XXXXXXXX")
    yield mktmp
    unless mktmp.success?
      error(@host.hostname, @t.step.install_pe.create_tmp_failed(error: mktmp.error_message(@host.hostname)))
      return
    end

    # TODO: Some parsing for the tmp dir? Define these better
    tmpdir = mktmp.results[@host.hostname][:stdout].strip
    installer_dir = "#{tmpdir}/install"
    answer_file = "#{tmpdir}/#{@host.hostname}.answers"

    chmodtmp = Installer::Action::Execute.new(:hosts => [@host], :cmd => "chmod 700 '#{tmpdir}'")
    yield chmodtmp
    unless chmodtmp.success?
      error(@host.hostname, @t.step.install_pe.chmod_tmp_failed(dir: tmpdir, error: chmodtmp.error_message(@host.hostname)))
      return
    end

    # Ensure log dir
    mklogdir = Installer::Action::Execute.new(:hosts => [@host], :sudo => true, :cmd => "mkdir -p #{@log_dir}")
    yield mklogdir
    unless mklogdir.success?
      error(@host.hostname, @t.step.install_pe.create_log_dir_failed(dir: @log_dir, error: mklogdir.error_message(@host.hostname)))
      return
    end

    # Chmod log dir
    logdir_chmod_cmd = "chmod 700 '#{@log_dir}'"
    chmodlogdir = Installer::Action::Execute.new(:hosts => [@host], :sudo => true, :cmd => logdir_chmod_cmd)
    yield chmodlogdir
    unless chmodlogdir.success?
      error(@host.hostname, @t.step.install_pe.chmod_log_dir_failed(dir: @log_dir, error: chmodlogdir.error_message(@host.hostname)))
      return
    end

    # Ensure /etc/puppetlabs/installer dir
    mketcdir = Installer::Action::Execute.new(:hosts => [@host], :sudo => true, :cmd => "mkdir -p #{@etc_dir}")
    yield mketcdir
    unless mketcdir.success?
      error(@host.hostname, @t.step.install_pe.create_etc_installer_failed(dir: @etc_dir, error: mketcdir.error_message(@host.hostname)))
      return
    end

    # Chmod installer dir
    installdir_chmod_cmd = "chmod 700 '#{@etc_dir}'"
    chmodinstalldir = Installer::Action::Execute.new(:hosts => [@host], :sudo => true, :cmd => installdir_chmod_cmd)
    yield chmodinstalldir
    unless chmodinstalldir.success?
      error(@host.hostname, @t.step.install_pe.chmod_etc_installer_failed(dir: @etc_dir, error: chmodinstalldir.error_message(@host.hostname)))
      return
    end

    # Ensure there is an installer available to copy
    unless File.readable? "#{@local_installer_dir}/puppet-enterprise-installer"
      error("localhost", @t.step.install_pe.installer_not_found(dir: @local_installer_dir))
      return
    end

    # Copy installer files
    installer = Installer::Action::Upload.new(:to_host => @host, :from => @local_installer_dir, :to => installer_dir, :opts => {:recursive => true})
    yield installer
    unless installer.success?
      error(@host.hostname, @t.step.install_pe.installer_not_copied(host: @host.hostname))
      return
    end

    # Copy answers file and save a copy locally
    answers = @host.generate_answers
    write_local_answers(@host.hostname, answers)

    answers = Installer::Action::Upload.new(:to_host => @host, :content => answers, :to => answer_file)
    yield answers
    unless answers.success?
      error(@host.hostname, @t.step.install_pe.installer_no_answers(host: @host.hostname))
      return
    end

    # Run installer with answer file
    install = Installer::Action::Execute.new(:host => @host, :sudo => true, :cmd => "#{installer_dir}/puppet-enterprise-installer -D -a #{answer_file}")
    yield install
    unless install.success?
      error(@host.hostname, @t.step.install_pe.installer_run_failed(host: @host.hostname))
      return
    end

    # Cleanup tmp if install succeeded. If not, we want to keep these files around
    # for troubleshooting purposes.
    cleanup = Installer::Action::Execute.new(:host => @host, :sudo => true, :cmd => "rm -rf '#{tmpdir}'")
    yield cleanup
    unless cleanup.success?
      error(@host.hostname, @t.step.install_pe.cleanup_tmp_failed(dir: tmpdir, host: @host.hostname, error: cleanup.error_message(@host.hostname)))
      return
    end
  end

  def write_local_answers(hostname, answers)
    Dir.mkdir(@local_answer_dir) unless File.directory? @local_answer_dir
    File.open("#{@local_answer_dir}/#{hostname}.answers", 'w') { |f| f.puts answers }
  end

  def fail_stop?
    true
  end

  def name
    "install_pe_#{@host}"
  end

  def description
    @t.step.install_pe.description(host: @host)
  end
end
